(*
 * The contents of this file are subject to the ICS(TM) Community Research
 * License Version 2.0 (the ``License''); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.icansolve.com/license.html.  Software distributed under the
 * License is distributed on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY
 * KIND, either express or implied. See the License for the specific language
 * governing rights and limitations under the License.  The Licensed Software
 * is Copyright (c) SRI International 2003, 2004.  All rights reserved.
 * ``ICS'' is a trademark of SRI International, a California nonprofit public
 * benefit corporation.
 *)
 

(** Datatype for storing variable disequalities *)
       

module Set = Set.Make(
  struct
    type t = Term.t * Jst.t
    let compare (x, _) (y, _) = Term.Var.compare x y
  end)

let mem x = 
  Set.exists (fun (y, _) -> Term.eq x y)

exception Found of Set.elt

(** Return [(x, rho)] if there is such an entry in [ds];
  otherwise raise [Not_found]. *)
let assoc x ds =
  try
    Set.iter
      (fun ((y, _) as d) ->
	 if Term.eq x y then 
	   raise(Found(d)))
      ds;
    raise Not_found
  with
      Found(d) -> d
	    

type t =  Set.t Term.Map.t

let empty = Term.Map.empty

let is_empty s = (s == empty)

let eq s t = (s == t)

let fold = Term.Map.fold
let iter = Term.Map.iter


(** All terms known to be disequal to [x]. *)
let diseqs s x =
  try Term.Map.find x s with Not_found -> Set.empty


(** Known disequalities; [x |-> {y, z}] is interpreted as [x <> y & x <> z].
  The function is closed in that forall [x], [y] such that [x |-> {..., y,...}]
  then also [y |-> {...., x,....}] *)
let closed s =
  Term.Map.fold
    (fun x ys acc -> acc && 
       (Set.for_all (fun (y, _) -> mem x (diseqs s y)) ys))
    s true

let occurs s x = Term.Map.mem x s


(** List all disequalities. *)
let to_set s =
  Term.Map.fold
    (fun x ys ->
       Set.fold 
         (fun (y, rho) -> 
	    let d = Fact.Diseq.make x y rho in
	      Fact.Diseq.Set.add d)
         ys)
    s 
    Fact.Diseq.Set.empty

let to_list s =
  Fact.Diseq.Set.fold (fun d acc -> d :: acc) (to_set s) []
	   


(** Pretty-printing. *)
let rec pp fmt s = 
  let ds = to_set s in
  if not(Fact.Diseq.Set.is_empty ds) then
    begin
      Pretty.list Fact.Diseq.pp fmt (Fact.Diseq.Set.elements ds)
    end
    


let map_diseqs s f a =
  let (x, rho) = f a in                        (* [rho |- x = a] *)
  let ds = diseqs s x in
    if Term.eq a x then ds else
      Set.fold 
	(fun (z, tau) ->                       (* [tau |- x <> z] *)
	   let sigma = Jst.dep2 tau rho in     (* ==> [sigma |- a <> z]. *)
	     Set.add (z, sigma))
	ds Set.empty

exception Found of Jst.t

(** Check if two terms are known to be disequal. *)
let is_diseq s x y =
  try
    Set.iter
      (fun (z, rho) ->
	 if Term.eq y z then
	   raise(Found(rho)))
      (diseqs s x);
    None
  with
      Found(rho) -> Some(rho)


(** Adding a disequality over variables *)
let add d s = 
  assert(Fact.Diseq.is_var d);
  assert(closed s);
  let (x, y, rho) = d in    (* [rho |- x <> y] *)
    match is_diseq s x y with
      | Some _ -> 
	  (s, Fact.Diseq.Set.empty)
      | None ->
	  Trace.msg "d" "Add(d)" d Fact.Diseq.pp;
	  let dx' = Set.add (y, rho) (diseqs s x)
	  and dy' = Set.add (x, rho) (diseqs s y) in
	  let s' = 
	    Term.Map.add x dx'
	      (Term.Map.add y dy' s)
	  in
	    assert(closed s');
	    (s', Fact.Diseq.Set.singleton d)

	
(** Propagating an equality between variables. *)
let merge e s =
  Trace.msg "d" "Merge(d)" e Fact.Equal.pp;
  assert(Fact.Equal.is_var e);
  assert(closed s);
  let x, y, rho = e in           (* [rho |- x = y] *)
    if Term.eq x y then 
      (s, Fact.Diseq.Set.empty)
    else
      match is_diseq s x y with
	| Some(tau) ->           (* [tau |- x <> y] *)
	    raise(Jst.Inconsistent(Jst.dep2 rho tau))
	| None -> 
	    let fresh = ref Fact.Diseq.Set.empty in
	    (try
	       let dx = Term.Map.find x s
	       and dy = diseqs s y in
		 assert(not(mem x dy));
		 assert(not(mem y dx));
		 let dx' = 
		   Set.fold 
		     (fun (z, tau) ->                       (* [tau |- x <> z] *)
			let sigma = Jst.dep2 tau rho in     (* ==> [sigma |- y <> z]. *)
			  fresh := Fact.Diseq.Set.add (Fact.Diseq.make y z sigma) !fresh;
			  Set.add (z, sigma))
		     dx Set.empty
		 in
		 let (s', dy') = 
		   (Term.Map.remove x s, Set.union dx' dy)   (* [dx] has to incorporate the new equality! *)
		 in
		 let (s'', dy'') = 
		   Set.fold
		     (fun (z, tau) (s, dy) ->                  (* [tau |- z <> x] *)
			let dz = diseqs s z in
			  try
			    let dz' = Set.remove (assoc x dz) dz in
			      assert(not(mem x dz'));
			      let sigma = Jst.dep2 tau rho in  (* [sigma|- z <> y] *)
			      let dz'' = Set.add (y, sigma) dz' in 
			      let dy' = Set.add (z, sigma) dy in
				fresh := Fact.Diseq.Set.add (Fact.Diseq.make y z sigma) !fresh;
				(Term.Map.add z dz'' s, dy')
			  with
			      Not_found -> (s, dy))
		     dx
		     (s', dy')
		 in
		 let s''' = 
		   if dy'' == dy then s'' else
		     Term.Map.add y dy'' s''
		 in
		   assert(closed s''');
		   assert(not(occurs s''' x));
		   (s''', !fresh)
	     with      
		 Not_found -> 
		   assert(not(occurs s x));
		   (s, !fresh))
 


(** Difference. *)
let diff d1 d2 =
  fold
    (fun x ds acc ->
       Set.fold 
       (fun (y, rho) acc ->
	  (match is_diseq d2 x y with
	     | Some _ -> acc
	     | None -> 
		 let (acc', _) = add (Fact.Diseq.make x y rho) acc in
		   acc'))
       ds acc)
    d1 empty


